/*
 * This file is part of WebLookAndFeel library.
 *
 * WebLookAndFeel library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * WebLookAndFeel library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with WebLookAndFeel library.  If not, see <http://www.gnu.org/licenses/>.
 */

package com.alee.extended.painter;

import javax.swing.*;
import java.awt.*;

/**
 * Simple color painter.
 * This painter simply fills component background with a single color.
 *
 * @param <E> component type
 * @author Mikle Garin
 * @see DefaultPainter
 * @see Painter
 * @since 1.4
 */

public class ColorPainter<E extends JComponent> extends DefaultPainter<E>
{
    /**
     * Color to fill component with.
     */
    protected Color color;

    /**
     * Constructs color painter with specified fill color.
     *
     * @param color color to fill component with
     */
    public ColorPainter ( Color color )
    {
        super ();
        this.color = color;
    }

    /**
     * Returns color to fill component with.
     *
     * @return color to fill component with
     */
    public Color getColor ()
    {
        return color;
    }

    /**
     * Sets color to fill component with.
     *
     * @param color new color to fill component with
     */
    public void setColor ( Color color )
    {
        this.color = color;
    }

    /**
     * Returns whether visual data provided by this painter is opaque or not.
     * Returned value might affect component opacity depending on painter support inside that component UI.
     *
     * @param c component to process
     * @return true if visual data provided by this painter is opaque, false otherwise
     */
    public boolean isOpaque ( E c )
    {
        return color != null && color.getAlpha () == 255;
    }

    /**
     * Paints visual data onto the component graphics.
     * Provided graphics and component are taken directly from component UI paint method.
     * Provided bounds are usually fake (zero location, component size) but in some cases it might be specified by componentUI.
     *
     * @param g2d    component graphics
     * @param bounds bounds for painter visual data
     * @param c      component to process
     */
    public void paint ( Graphics2D g2d, Rectangle bounds, E c )
    {
        // Do not paint anything if color is not set
        if ( color != null )
        {
            // Determining actual rect to be filled (we don't need to fill invisible area)
            Rectangle r = c.getVisibleRect ().intersection ( bounds );

            // If there is anything to fill we do it
            if ( r.width > 0 && r.height > 0 )
            {
                g2d.setColor ( color );
                g2d.fillRect ( r.x, r.y, r.width, r.height );
            }
        }
    }
}